IMPLEMENTATION MODULE BMP;

(* Some operations on BMP graphics files. *)
(* This module is only for 16 bit compilers on a 16 bit operating system. *)


(* V4.0B M. v/d Voort Jul./27/95 - TS 3.1 version (TS 1.17 will probably work
                                    too) qpv can read the binaries generated
                                    by this version.
   V4.0, J.Andrea, Jan.2/95 -FST M2 version, no separate header, corrected
                             used and important colors numbers, padding
                             operation corrected, and counters are 16 bit.
   V3.0, J.Andrea, Feb.8/94 -define all pixel sizes
   V2.0, J.Andrea, Jan.30/94
     this code may be freely used and distributed, it may not be sold *)

IMPORT FIO;
FROM Storage IMPORT ALLOCATE, DEALLOCATE;

CONST
   max_per_byte = 8;   (* maximum number of pixels per byte *)
   zero = LONGCARD(0);          (* long zero *)

TYPE
  FileHeader = RECORD
                 type :ARRAY [1..2] OF CHAR;
                 size :LONGCARD;
                 reserved :ARRAY [1..2] OF CARDINAL;
                 offset :LONGCARD;
               END;

  InfoHeader = RECORD
                 size :LONGCARD;
                 width, height :LONGCARD;
                 planes :CARDINAL;
                 bit_count :CARDINAL;
                 compression :LONGCARD;
                 image_size :LONGCARD;
                 x_per_meter, y_per_meter :LONGCARD;
                 colors_used, colors_important :LONGCARD;
               END;

  ColorMapItem = RECORD
                   red, green, blue, empty :BYTE;
                 END;

  Pixels      = RECORD
                  out :CARDINAL;    (* size of current scan line *)
                  ready :CARDINAL;
                  count :CARDINAL;
                  max_value :CARDINAL;
                  values :ARRAY [1..max_per_byte] OF CARDINAL; (* possible
values that fit one byte *)
                  multiply :ARRAY [1..max_per_byte] OF CARDINAL;
                  pad_size :CARDINAL;
                END;

  BMPFile = POINTER TO RECORD
               disk   :FIO.File;
               file   :FileHeader;
               info   :InfoHeader;
               pixels :Pixels;
            END;



  (* ------------------------------------------------- *)
  PROCEDURE Create( VAR bmp :BMPFile; name :ARRAY OF CHAR;
          height, width :CARDINAL;
          bits_per_pixel :CARDINAL;
          used_colors, important_colors :CARDINAL;
            reds, greens, blues :ARRAY OF CARDINAL;
          VAR error :CARDINAL );

  VAR
    n_written      :CARDINAL;
    bytes_per_line :CARDINAL;
    image_size     :LONGCARD;
    i              :CARDINAL;
    done           :BOOLEAN;

    (* --------------------------------------- *)
    PROCEDURE SetupFileHeader;
    VAR
      headers_size :CARDINAL;
    BEGIN

      IF bits_per_pixel # 24 THEN

        (* the max value + 1 is really the number of colors *)
        headers_size := ( bmp^.pixels.max_value + 1 ) * SIZE(ColorMapItem)
                        + SIZE(FileHeader) + SIZE(InfoHeader);

      ELSE
        (* a 24 bit image needs no palatte *)
        headers_size := SIZE(FileHeader) + SIZE(InfoHeader);
      END;

      (* the first two bytes (magic number) are required to be 'BM' *)
      bmp^.file.type[1] := 'B';  bmp^.file.type[2] := 'M';

      (* the size of the file is the size of the 4 sections *)
      bmp^.file.size := LONGCARD( headers_size ) + image_size;

      (* reserved values always set to zero *)
      bmp^.file.reserved[1] := 0;  bmp^.file.reserved[2] := 0;

      (* the offset is really just the size of the headers *)
      bmp^.file.offset := LONGCARD( headers_size );

    END SetupFileHeader;

    (* --------------------------------------- *)
    PROCEDURE SetupInfoHeader;
    BEGIN
      (* this item means, the size of this header *)
      bmp^.info.size := LONGCARD( SIZE(InfoHeader) );

      (* image dimensions as passed to this procedure *)
      bmp^.info.width  := LONGCARD( width );
      bmp^.info.height := LONGCARD( height );

      (* number of planes must be set to one *)
      bmp^.info.planes := 1;

      (* number of bits per pixel *)
      bmp^.info.bit_count := bits_per_pixel;

      (* no compression, choose zero *)
      bmp^.info.compression := zero;

      (* size of the image in bytes, most applications set this to zero,
         i'm going to use the true size *)
      bmp^.info.image_size := image_size;

      (* don't do anything with pixel sizes *)
      bmp^.info.x_per_meter := zero;  bmp^.info.y_per_meter := zero;

      (* size of the color table *)
      IF bits_per_pixel = 24 THEN
        (* true colors have no color table *)
        bmp^.info.colors_used := zero;
      ELSE
        (* most applications just use zero here, and then the readers can
           compute the number of extries from the size of the table. *)
        bmp^.info.colors_used := LONGCARD( used_colors );
      END;

      (* number of important colors, the displayer can use this to help
         pick the best palatte *)
      IF bits_per_pixel = 24 THEN
        (* all are important, choose zero *)
        bmp^.info.colors_important := zero;
      ELSE
        (* most applications just user zero here *)
        (* for now i'll to that too *)
        bmp^.info.colors_important := zero;
      END;

    END SetupInfoHeader;

    (* --------------------------------------- *)
    PROCEDURE TestColors( VAR done :BOOLEAN );
    VAR
      i, n_colors :CARDINAL;
    BEGIN

      (* only possible values are 1,4,8,24. Truncate to lower value *)

      (* the multiplier will be used later to shift values into the
         proper locations in the output byte *)
      bmp^.pixels.multiply[1] :=  1;

      IF bits_per_pixel < 4 THEN

        bits_per_pixel    := 1;
        n_colors          := 2;
        bmp^.pixels.count := 8;
        FOR i := 2 TO max_per_byte DO
          bmp^.pixels.multiply[i] := 2 * bmp^.pixels.multiply[i-1];
        END;

      ELSIF bits_per_pixel < 8 THEN

        bits_per_pixel    := 4;
        n_colors          := 16;
        bmp^.pixels.count := 2;
        bmp^.pixels.multiply[2] := 16;

      ELSIF bits_per_pixel < 24 THEN

        bits_per_pixel    := 8;
        n_colors          := 256;
        bmp^.pixels.count := 1;

      ELSE

        bits_per_pixel    := 24;
        n_colors          := 256;  (* really 2^24, but pick any number *)
        bmp^.pixels.count := 1;

      END;

      bmp^.pixels.max_value := n_colors - 1;

      (* the number of used colors can't be any larger than the
         possible number of colors *)
      IF used_colors > n_colors THEN used_colors := n_colors END;

      (* the number of important colors can't be any larger than the
         possible number of used colors *)
      IF important_colors > used_colors THEN important_colors := used_colors
END;

      (* test sizes of arrays too *)
      (* since the lower array bound is 0, the upper must be at least the
         number of values that a pixel can take on (minus one) *)

      IF bits_per_pixel # 24 THEN
        IF HIGH( reds ) < used_colors - 1 THEN
          done  := FALSE;
          error := 3;
        ELSE
          IF HIGH( greens ) < used_colors - 1 THEN
            done  := FALSE;
            error := 4;
          ELSE
           IF HIGH( blues ) < used_colors - 1 THEN
             done  := FALSE;
             error := 5;
           END;
          END;
        END;
      END;

    END TestColors;

    (* -------------------------------------- *)
    PROCEDURE BytesPerLine;
    VAR
      n, bits_per_line :CARDINAL;
    BEGIN

      (* bmp files need to be padded to multiples of 4 bytes *)
      (* compute the number of bytes needed to do the padding *)

      (* unfortualtely an overflow may occur if the file is very wide *)

      IF bits_per_pixel = 24 THEN
        bytes_per_line := width * 3;

      ELSIF bits_per_pixel = 8 THEN
        bytes_per_line := width;

      ELSE

        bits_per_line := width * bits_per_pixel;

        (* round that value up to a multiple of 8, so that an integral
           number of bytes is filled *)
        n := bits_per_line MOD 8;
        IF n # 0 THEN
          bits_per_line := bits_per_line + 8 - n;
        END;

        bytes_per_line := bits_per_line DIV 8;

      END;

      (* round up the number of bytes per line to a multiple of 4 *)

      n := bytes_per_line MOD 4;
      IF n = 0 THEN
        bmp^.pixels.pad_size := 0;
      ELSE
        bmp^.pixels.pad_size := 4 - n;
      END;
      (* and add to the bytes per line to take care of the padding too *)
      bytes_per_line := bytes_per_line + bmp^.pixels.pad_size;

      (* also, determine the number of bytes that will be needed to
         create the image section in the output file *)
      image_size := LONGCARD( bytes_per_line ) * LONGCARD( height );

    END BytesPerLine;

    (* -------------------------------------- *)

  BEGIN

    done  := TRUE;
    error := 0;

    NEW( bmp );
    bmp^.disk:=FIO.Create(name);
    IF NOT (FIO.IOresult()=0) THEN
      error := 1;
      DISPOSE( bmp );
    ELSE
      IF ( height = 0 ) OR ( width = 0 ) THEN
        done := FALSE; error := 2;
        DISPOSE( bmp );
      ELSE

        TestColors( done );
        IF done THEN

          BytesPerLine;

          (* based on a file of the given dimensions, compute the remainder of
             the parameters *)

          SetupFileHeader;
          FIO.WrBin( bmp^.disk,bmp^.file, SIZE(FileHeader));

            SetupInfoHeader;
            FIO.WrBin( bmp^.disk, bmp^.info ,SIZE(InfoHeader));

              (* output the color table, if needed *)
              (* i don't do any i/o checking here for no good reason *)

              IF bits_per_pixel # 24 THEN
                FOR i := 0 TO used_colors - 1 DO
                   FIO.WrChar(bmp^.disk,CHR( blues[i] )); FIO.WrChar(bmp^.disk,CHR( greens[i] ));
                   FIO.WrChar(bmp^.disk,CHR( reds[i] ));  FIO.WrChar(bmp^.disk,CHR( 0 ));
                END;
              END;

              (* initialize the output pixel counters *)

              bmp^.pixels.out   := 0;
              bmp^.pixels.ready := 0;
        END;
      END;
    END;
  END Create;

  (* -------------------------------------------------- *)
  PROCEDURE Close( VAR bmp :BMPFile );
  VAR
    done :BOOLEAN;
  BEGIN
    FIO.Close( bmp^.disk );
    DISPOSE( bmp );
  END Close;


  PROCEDURE WritePixel( bmp :BMPFile; r, g, b :CARDINAL );
  VAR
    i :CARDINAL;
    out_value :CARDINAL;

  BEGIN

    INC( bmp^.pixels.out );

    IF bmp^.info.bit_count = 24 THEN

      (* note that the ordering is reversed *)
      FIO.WrChar(bmp^.disk,CHR( b ));
      FIO.WrChar(bmp^.disk,CHR( g ));
      FIO.WrChar(bmp^.disk,CHR( r ));

    ELSIF bmp^.info.bit_count = 8 THEN

      FIO.WrChar(bmp^.disk,CHR( r ));

    ELSE

      INC( bmp^.pixels.ready );

      IF r > bmp^.pixels.max_value THEN
        bmp^.pixels.values[bmp^.pixels.ready] := bmp^.pixels.max_value;
      ELSE
        bmp^.pixels.values[bmp^.pixels.ready] := r;
      END;


      (* for the pixel sizes that are smaller than one byte,
         there are two conditions that will cause the current byte to
         be completed
         1) enough pixels have been setup to fill a byte
         2) enough pixels have been available to fill a scan line *)

      IF ( bmp^.pixels.ready     = bmp^.pixels.count )
      OR ( LONGCARD(bmp^.pixels.out) = bmp^.info.width ) THEN

        (* write those values into a byte and output that byte *)
        (* the multiplier will push the value over the correct number of
           bit spaces into the corresponding position in the output byte *)
        out_value := 0;

        (* only run to the ready number, in case its the end of line condition
           that has been reached *)
        FOR i := 1 TO bmp^.pixels.ready DO
          out_value := out_value +
                       bmp^.pixels.multiply[i] * bmp^.pixels.values[i];
        END;
        FIO.WrChar(bmp^.disk,CHR( out_value ));

        (* reset to none ready *)
        bmp^.pixels.ready := 0;
      END;

    END;

    (* if this scan line is finished, then fill it out *)
    IF LONGCARD( bmp^.pixels.out ) = bmp^.info.width THEN
      FOR i := 1 TO bmp^.pixels.pad_size DO
        FIO.WrChar(bmp^.disk,CHR( 0 ));
      END;
      (* reset to zero output *)
      bmp^.pixels.out := 0;
    END;

  END WritePixel;

END BMP.


