MODULE Dictionary;
(* Dictionary, part of the XtdLib package. (C) Marco van de Voort 1997

  Contains a VID bug. Working on it DO NOT RUN WHEN COMPILED VID=OFF.
  It might go good, but some textfiles (30-50%) hang your computer, so
  tightly that even Win95 crashes.

  Based on a SWAG source, so free for none commercial use.

  Original (SWAG author, UNKNOWN) comments:
  simple dictionary using a btree.
  The program reads in an ASCII file with one word per line and stores the
  words in an btree. A btree is something like binary tree but every node
  can have more than two descent nodes. This is done by linked list.

  This method has two advantages:
    * when a word is wrong you can easily give some proposes how the word
      is written correctly (just change the path in the tree a little)
    * bigger dict. may save space. E.g "base, basicly, basement" etc.
      share the same path on the first three niveaus.

  ATTENTION! I don't free any mem I've allocated. This is done by the
  heap manager (i.e. he allocates large blockes and releases them ) when
  the program ends. But this can be added easily.

  Also, there is no PROCEDURE included that deletes words (I don't need it in
  my project). I suggest it is NOT that easy TO add such a PROCEDURE but
  have a try ;-))    *)

(* MvdV comments:
   -  It is a binary list isn't it? Ok, using the node right is considered
       to be the same level as the node itself. But that's a point of view?
   -  Space save argument is nonsens. Two pointers per char!(8 bytes!)
      A tokenized matching system is much better(as far as space is concerned!)
      The space difference of Take and Taken when written as a char is 1,
      with this binary system you save 4, and the extra char "N" costs 10
      bytes. And these word are very alike! Most words only match the first
      or first two characters! (effectively even 16 bytes, because the memory
      is allocated in paragraphs!)
      Storing all two character pairs costs 10*26*26=6760 bytes
      Storing them in a pack char array costs 26*3 (2 chars, 1 char separator)
      =78 bytes. Almost 100* as little!
   -  Instead of a nicely textfile, with input on each line, a standard
      textfile is used. I use XTDLIB.DOC as a test. Hmm, maybe I should use
      4dos.doc next :-)
      I tried 00global.txt (800k, index simtel CD set), but it went out of
      memory.
   -  This program has no output! So I added an output procedure.
      The dictionary is spooled to a textfile, with a word per line.
      It's sorted because the list is kept sorted.
   -  The original didn't implement building a sorted list, allthough it was
      only 4 extra statements or so. I did :-)
   -  Added global deallocate.
   -  DeleteWord was more difficult than I thought, but not as much as
       the original author claims. I have a simpler approach, but the
       current approach doesn't use temporary variables arrays.
       (original idea was scanning for the word, and saving a pointer for
        each character. The size of the array would however limit the size
        of the word, so I choosse this way algoritm)
    Rough estimation of max words : 5000-10000 depending on "double" words.
     It's a demo, not an application, requires an extender to work properly.
     Memory usage is too heavy.
   -  Memory allocation seems to work, afterwards all bytes free again.

How to use it? Make a list of all your textfiles per language, path inclusive.
Run Dictionary with one of the files as a parameter. The system
loads your first file into memory, and then you type
templ <filename>

with <filename>=the textfile you created.

Now type "end", and your rough dictionary is in newdict.
   *)

IMPORT FIO,IO;
FROM Str  IMPORT Length,Caps,Compare,ItemS,CHARSET,Delete,Copy;
FROM Lib  IMPORT EnableBreakCheck,ParamCount,ParamStr;
FROM ELib IMPORT ReplaceChar,SetCharPos,KillBChar,KillChar,StripChar;
FROM Storage IMPORT ALLOCATE,DEALLOCATE,Available,HeapTotalAvail,MainHeap;

CONST
      CrLf         = CHR(13)+CHR(10);
      (* Displayed when HELP is typed in interrogation mode *)
HelpMenu='END               : end interrogation mode.'+CrLf+
         'DEL <word>        : delete word <word>'+CrLf+
         'HELP              : this menu'+CrLf+
  'LOAD <File>       : merges file <filename> with current dictionary'
         +CrLf+
   'TEMPL <File>      : merges data of (several) files in textfile <File>'
         +CrLf+
         'OUTPUTFILE <File> : Set name of output file to <File>'+CrLf+
         'SAVE <File>       : Save all data to the file'+CrLf+
         '<word>            : Does <word> exist?'+CrLf+ CrLf;

     (* The main type. 10 bytes, but 16 are allocated. *)
TYPE  PNode     = POINTER TO TNode;
      TNode     = RECORD
                    Character : CHAR;    (* the current Character *)
                    WordEnd   : BOOLEAN; (* is this char. the last of one word?*)
                    right,down: PNode;   (* right: points to next char on the
                                                  same niveau
                                           down : points to the next char in
                                                  word *)
                  END;


VAR BTree: PNode;                       (* our tree *)

PROCEDURE Memo;
BEGIN
 IO.WrStr("Memory available :");
 IO.WrLngCard(LONGCARD(HeapTotalAvail(MainHeap))*16,15); IO.WrLn;
END Memo;

PROCEDURE CreateBTree;
(* just initalizes the tree w/ a dummy element *)
BEGIN
  BTree:=NIL;
  NEW(BTree);
  BTree^.Character:=CHR(1AH); (* #$1A is END-OF-FILE. shouldn't be used in any word *)
  BTree^.right:=NIL;
  BTree^.down:=NIL;
  BTree^.WordEnd:=TRUE;
  (* The dummy element can be deleted once words have been added. It isn't
  done yet *)
END CreateBTree;

PROCEDURE GetNode(Character: CHAR; LevelPtr: PNode): PNode;
(* returns the node in Level "LevelPtr" that contains "Character".
  if there is no node, it is created, in an alphabetical way. *)
VAR P,p2: PNode;

BEGIN (* All allocation is done in this procedure, so one check is enough*)
  IF NOT Available(SIZE(TNode)) THEN (* Safety measure. Don't delete, the
                                         system will lock up very tight, and
                                         even crash Windows 95!*)
    IO.WrStr("Out of Memory, trying to spool memory would crash system. So bye bye!");
    IO.WrLn;
    HALT;
    END;
  IF LevelPtr=NIL THEN (* New level*)
    NEW(P);
    P^.right:=NIL;
    P^.down:=NIL;
    P^.Character:=Character;
    P^.WordEnd:=FALSE;
    RETURN P;
   ELSE    (* Search for a pointer in this level, if not found, create
              a new one.*)
    P:=LevelPtr;
    WHILE (P^.right<>NIL) AND (P^.Character<>Character) DO
     P:=P^.right;
     END;
    IF P^.Character=Character THEN (* Found! *)
      RETURN P;
      ELSE
       p2:=LevelPtr;    (* Search level again*)
       P:=p2;
       WHILE (p2<>NIL) AND (p2^.Character<Character) DO
        (* Search for a place to insert*)
        P:=p2;
        p2:=p2^.right
        END;
       IF p2=P THEN             (* Maybe insert before first pointer,
                                   or only one pointer*)
         NEW(p2);               (* Create a new pointer*)
         p2^.right:=P^.right;   (* link it in *)
         P^.right:=p2;
         IF P^.Character<Character THEN (* Insert after 1st pointer*)
          P:=p2;
         ELSE
          p2^.Character:=P^.Character; (* Insert before 1st pointer*)
          p2^.WordEnd:=P^.WordEnd;     (* move record*)
          p2^.down:=P^.down;
          END;
         P^.Character:=Character;       (* add new data to "new" record*)
         P^.down:=NIL;
         P^.WordEnd:=FALSE;
        ELSE
         p2:=P^.right; (* Save link *)
         NEW(P^.right);   (* Create new link *)
         P:=P^.right;     (* and follow it *)
         P^.right:=p2;    (* Make link to next record *)
         P^.Character:=Character; (* rest of record*)
         P^.down:=NIL;
         P^.WordEnd:=FALSE; (* Is set by calling procedure if necessary.*)
         END;

        (* p^.right is NIL! *)

       RETURN P; (* Return the (newly created or found) pointer *)
      END; (*if*)
  END; (* if *)
END GetNode;

PROCEDURE InsertWord(wort: ARRAY OF CHAR);
(* inserts the word "wort" into btree *)
VAR P1,p2,p3: PNode;
    i: CARDINAL;
BEGIN
  IF (wort[0]=CHR(0)) OR (wort[1]=CHR(0)) THEN RETURN; END;
  p2:=BTree;
  FOR i:=0 TO Length(wort)-1 DO
    P1:=GetNode(wort[i],p2);        (* Search for pointer on this level*)
    IF P1=NIL THEN
     RETURN;                        (* Something's wrong. Original meant
                                       for post-mortem dump, but I keep it
                                       as a safety measure*)
     END;
    IF p2=NIL THEN                  (* Make link to last level *)
     p3^.down:=P1;
     END;
    p3:=P1;
    p2:=P1^.down;                   (* Search next level*)
  END;
  P1^.WordEnd:=TRUE;
END InsertWord;

PROCEDURE ProofWord(Wort: ARRAY OF CHAR): BOOLEAN;
(* returns true if "wort" is in our dictionary *)
VAR P1,p2: PNode;
    I: CARDINAL;
BEGIN
  IF Wort[0]=CHR(0) THEN RETURN FALSE; END;
  P1:=BTree;
  I:=0;
  WHILE (P1<>NIL) AND (Length(Wort)>=I) DO
    WHILE (P1^.right<>NIL) AND (P1^.Character<>Wort[I]) DO P1:=P1^.right; END;
    IF P1^.Character=Wort[I] THEN
      INC(I);
      p2:=P1;
      P1:=P1^.down;
    ELSE P1:=NIL; END;
  END;
  IF (I=Length(Wort)) AND (p2^.WordEnd) THEN RETURN TRUE; END;
  RETURN FALSE;
END ProofWord;

PROCEDURE KillWord(Woord : ARRAY OF CHAR);
(* Based on Proofword. Runs a modified version of proofword to find last
   character. If last character has no links, then last characters record
   is deleted. The last character is also deleted from the string, and the
   search for the (new) last character is repeated *)

VAR Len : CARDINAL;
    P1,P2,P3: PNode;
    I    : CARDINAL;

BEGIN
 Caps(Woord);
 Len:=Length(Woord);    (* Save original length word *)
 LOOP
  IF Woord[0]=CHR(0) THEN  (* All character scanned? *)
   RETURN;
   END;
  P1:=BTree;               (* Start at beginning of tree *)
  I:=0;
  WHILE (P1<>NIL) AND (Length(Woord)>=I) DO (* Repeat until end of word *)
    WHILE (P1^.right<>NIL) AND (P1^.Character<>Woord[I]) DO
    (* search level for matching character*)
     P1:=P1^.right;
     END;
    IF P1^.Character=Woord[I] THEN (* Found on this level? Next level*)
      INC(I);
      P3:=P2;
      P2:=P1;
      P1:=P1^.down;
     ELSE
      P1:=NIL (* Not found of end of word*)
     END;
   END;
  IF I=Length(Woord) THEN (* Entire word found?*)
   IF I=Len THEN
     P2^.WordEnd:=FALSE; (* If last character we're deleting? Then WordEnd
                              becomes false since the word no longer exists*)
     END;
    IF (P2^.right#NIL) OR (P2^.down#NIL) THEN (* Links to other (parts) of
                                                 words?*)
     RETURN; (* Yes, stop deleting *)
     END;
    IF P2^.WordEnd = TRUE THEN       (* another word stops here? *)
     RETURN; (* Yes, have to keep this record. Stop deleting*)
     END;
    P3^.down:=NIL;                   (* Delete reference to current level
                                         (in P2) *)
    DISPOSE(P2);                     (* Kill current level*)
    Woord[I-1]:=CHR(0);              (* Remove last character of word*)
   ELSE
    RETURN;
    END;
  END;
END KillWord;

PROCEDURE KillTree(R:PNode);
(* Recursive, delete a binary tree. All nodes except Root-node are deleted.
   Root-node can't be killed because of recursion.
   Requires extra procedure calling this one. Not worth the trouble.
   *)

BEGIN
 LOOP
  IF R^.right # NIL THEN
   IF (R^.right^.right = NIL) AND (R^.right^.down=NIL) THEN
    DISPOSE(R^.right);
   ELSE
    KillTree(R^.right);
    END;
   END;
  IF R^.down # NIL THEN
   IF (R^.down^.right = NIL) AND (R^.down^.down=NIL) THEN
    DISPOSE(R^.down);
   ELSE
    KillTree(R^.down);
    END;
   END;
  IF (R^.right=NIL) AND (R^.down=NIL) THEN
   EXIT
   END;
  END;
END KillTree;

VAR Deep   : CARDINAL;              (* Recursionlevel, and index in OutStr*)
    OutStr : ARRAY[0..100] OF CHAR; (* Words are "saved" in here *)
    DictFile: FIO.File;             (* our ascii dictionary *)

PROCEDURE Output(R:PNode);
(*Recursive output of words to file. *)
BEGIN
 IF Deep>100 THEN               (* Safety, don't write behind string *)
  Deep:=2;
  END;
 OutStr[Deep]:=R^.Character; (* Save character of current node *)
 IF R^.WordEnd THEN          (* End of word? Then output! *)
  OutStr[Deep+1]:=CHR(0);
  FIO.WrStr(DictFile,OutStr);
  FIO.WrLn(DictFile);
  END;
 IF R^.down = NIL THEN      (* No "longer" words, or more characters in
                               this word?*)
  IF R^.right # NIL THEN    (* No words starting to differ on this point?*)
   Output(R^.right);
   END;
 ELSE
  INC(Deep);                (* Deeper words*)
  Output(R^.down);
  DEC(Deep);
  IF R^.right # NIL THEN    (* Longer words*)
   Output(R^.right);
   END;
  END;
END Output;

VAR  Buf   : ARRAY[0..4095] OF BYTE;

 (* Trim these character of both sides*)
CONST KillSet=CHARSET{'#',CHR(39),'?','0'..'9','/','-',':','\','*','"',
                      '$','&','+','<','>','=',';','@','^','|','~',']',
                      '[','_',CHR(123),CHR(125),'!'};

(* Don't scan words containing the following characters (after trimming)*)
      DontScan=CHARSET{CHR(2),'\','%','/',CHR(39),'0'..'9','$','&','"'};

      ItemString="_+. ()-:=,";    (* Separators used *)

      (* Possible commands in interrogation mode*)
      CommandStr=   'END@LOAD@DEL@HELP@OUTPUTFILE@TEMPL@ADD@SAVE';

 InitTxt='Dictionary v0.02, by Marco van de Voort (but based on a SWAG source)';
 Usage  ='A textfile as input is required, but is only read, not written'+CrLf;

PROCEDURE LoadOrMergeFile(FileName:ARRAY OF CHAR);
(* Load a textfile, split lines to words, and add words to the tree*)

VAR A,Line:CARDINAL;
    S,S2 : ARRAY[0..100] OF CHAR;

BEGIN
  DictFile:=FIO.Open(FileName);
  FIO.AssignBuffer(DictFile,Buf);
  IO.WrStr('Loading/Merging file '); IO.WrStr(FileName);
  IO.WrStr(' dictionary...');        IO.WrLn;
  A:=0;
  Line:=0;
  LOOP
   FIO.RdStr(DictFile,S);       (* read a line *)
   IF FIO.EOF THEN
    EXIT
    END;
   INC(Line);
   IO.WrCard(Line,5);
   ReplaceChar(S,"!"," ");      (* Remove exclamation marks*)
   Caps(S);
   A:=0;
    LOOP
     ItemS(S2,S,ItemString,A); (* Split into rough words*)
     IF S2[0]=CHR(0) THEN
      EXIT;
      END;
     StripChar(S2,'"');         (* Trim words*)
     KillChar(S2,KillSet);
     KillBChar(S2,KillSet);
     IF SetCharPos(S2,DontScan)=MAX(CARDINAL) THEN
       (* SetCharPos prevents words like "I'll" to be added to tree. *)
      InsertWord(S2);
      END;
     INC(A);
    END;
  END;
  FIO.Close(DictFile);
  IO.WrStr('done.'); IO.WrLn;
  Memo;
END LoadOrMergeFile;

VAR OutName:ARRAY[0..100] OF CHAR;

PROCEDURE OutWithIt;
(* Output file*)

BEGIN
   DictFile:=FIO.Create(OutName);
   FIO.AssignBuffer(DictFile,Buf);
   Deep:=0; (* Very important. Can't be done in Output because of recursion!*)
   Output(BTree^.right); (* Don't use the dummy record*)
   FIO.Close(DictFile);
END OutWithIt;

VAR S,S2,
    S3  : ARRAY[0..100] OF CHAR;
    A,B : CARDINAL;
    F   : FIO.File;
    Fl  : BOOLEAN;


BEGIN
  Memo;
  IO.WrStr(InitTxt); IO.WrLn;
  OutName:="newdict.doc";
  IF ParamCount()<1 THEN
   IO.WrStr(Usage);      (* No parameters*)
   HALT;
   END;
  ParamStr(S,1);
  EnableBreakCheck;
  CreateBTree;           (* Set up tree*)
  IF FIO.Exists(S) THEN
   LoadOrMergeFile(S);  (* Load file *)
  ELSE
   IO.WrStr(Usage);
   HALT;
   END;
  IO.WrStr('Dictionary maintenance mode. END WITH "END"!'); IO.WrLn;
  S:='';
  LOOP
    IO.WrStr('OK>');   (* Interrogation mode *)
    IO.RdStr(S);
    Caps(S);
    ItemS(S2,S,' ',0); (* Get 1st substring separated by spaces*)
    A:=0; Fl:=FALSE;
    REPEAT          (* Compare the substring with the commands*)
     ItemS(S3,CommandStr,'@',A);
     IF Compare(S2,S3)=0 THEN
      Fl:=TRUE             (* Found!*)
     ELSE
      INC(A);
      END;
    UNTIL Fl OR (Length(S3)=0);
    IF Fl AND (S3[0]#CHR(0)) THEN
           (* A is index in commandline *)
     Delete(S,0,Length(S2)+1); (* Determine argument,
                                Item(S2,s,[' '],1); is also possible *)
     CASE A OF
       0 : EXIT;                         (* END*)
      |1 : IF FIO.Exists(S) THEN          (* LOAD*)
            LoadOrMergeFile(S)
           ELSE
            IO.WrStr('File not found'); IO.WrLn;
            END;
      |2 :  KillWord(S);                  (* DEL*)
      |3 :  IO.WrStr(HelpMenu);
      |4 :  Copy(OutName,S);                   (*OutputName*)
      |5 :  IF FIO.Exists(S) THEN         (*TEMPL*)
             IO.WrStr("Merging files in ");
             F:=FIO.Open(S);
             LOOP
              FIO.RdStr(F,S);
              IF FIO.EOF THEN
               EXIT
               END;
              IF FIO.Exists(S) THEN
               LoadOrMergeFile(S);
               END;
               END;
              FIO.Close(F);
             END;
       |6:                                 (* ADD *)
            Caps(S);
            InsertWord(S);
       |7 :                                (*SAVE*)
            IO.WrStr('Saving file ');IO.WrStr(S); IO.WrLn;
            Copy(OutName,S);
            OutWithIt;
        END;
      A:=MAX(CARDINAL);
    ELSE          (* No keyword? Test of word exists in Tree*)
     IO.WrStr('WORD '); IO.WrStr(S);
     IF NOT ProofWord(S) THEN
      IO.WrStr(' NOT ');
      END;
     IO.WrStr('FOUND!'); IO.WrLn;
     END;
    END;   (* Go on forever, the exit wil do, saves a boolean:-) *)

   OutWithIt;                   (* Output tree to textfile*)
   KillTree(BTree);             (* Remove tree from memory*)
   DISPOSE(BTree);              (* Remvoe root-node*)
   Memo;
END Dictionary.